"""
Input datafiles for tests
"""

__all__ = [
        "PSF_FRAME_ZERO_PKA",
        "PSF_FRAME_NINETY_PKA",
        "INDEXERR_FRAME14_GRO",
        "INDEXERR_FRAME14_XTC",
        "ATTERR_FRAME1_PDB",
        "ATTERR_FRAME1_XTC",

]

from pkg_resources import resource_filename

PSF_FRAME_ZERO_PKA = resource_filename(__name__,
                                       './data/psf_ref_frame0.pka')
PSF_FRAME_NINETY_PKA = resource_filename(__name__,
                                         './data/psf_ref_frame90.pka')
# Issue #10
INDEXERR_FRAME14_GRO = resource_filename(__name__,
                                         './data/indexerr_frame14.gro')
INDEXERR_FRAME14_XTC = resource_filename(__name__,
                                         './data/indexerr_frame14.xtc')
ATTERR_FRAME1_PDB = resource_filename(__name__,
                                      './data/atterr_frame1.pdb')
ATTERR_FRAME1_XTC = resource_filename(__name__,
                                      './data/atterr_frame1.xtc')



"""
Tests that propkatraj can be imported.
"""

import os

import MDAnalysis as mda
import numpy as np
import pandas as pd
import pytest

from MDAnalysisTests.datafiles import PSF, DCD, GRO, XTC
from numpy.testing import assert_almost_equal, assert_equal

import propkatraj
from propkatraj.tests.datafiles import (PSF_FRAME_ZERO_PKA,
                                        PSF_FRAME_NINETY_PKA,
                                        INDEXERR_FRAME14_GRO,
                                        INDEXERR_FRAME14_XTC,
                                        ATTERR_FRAME1_PDB,
                                        ATTERR_FRAME1_XTC,)


@pytest.fixture(scope='module')
def u():
    return mda.Universe(PSF, DCD)


@pytest.fixture(scope='module', params=(
    (INDEXERR_FRAME14_GRO, INDEXERR_FRAME14_XTC),
    (ATTERR_FRAME1_PDB, ATTERR_FRAME1_XTC)
))
def u_badframe(request):
    top, traj = request.param
    return mda.Universe(top, traj)


@pytest.fixture(scope='module')
def glu_ref():
    arr = np.array([2.53, 2.61, 2.84, 2.57, 2.76, 2.51, 2.68, 2.95,
                    3.18, 3.34])
    return arr


def pka_from_file(filename):
    columns = ['resname', 'resnum', 'chain', 'pka', 'model-pka']
    df = pd.read_csv(filename, names=columns, skiprows=1,
                     delim_whitespace=True)
    df.sort_values(by=['resnum'], inplace=True)
    resnums = df['resnum'].to_numpy()
    pkas = df['pka'].to_numpy()
    return resnums, pkas


@pytest.mark.parametrize('selection', ['ag', 'default'])
def test_single_frame_regression(tmpdir, u, selection):
    """Single frame PropkaTraj regression test"""
    with tmpdir.as_cwd():
        if selection == 'ag':  # test slection by passing an atomgroup
            pkatraj = propkatraj.PropkaTraj(u.select_atoms('protein'))
        else:  # test selection by using the default select="protein"
            pkatraj = propkatraj.PropkaTraj(u)

        pkatraj.run(stop=1)

        # load reference data
        ref_resnums, ref_pkas = pka_from_file(PSF_FRAME_ZERO_PKA)

        # test residue numbers
        resnums = pkatraj.results.pkas.columns.to_numpy()
        assert_equal(resnums, ref_resnums)

        # test pka values
        assert_almost_equal(pkatraj.results.pkas.values[0], ref_pkas,
                            decimal=2)


def test_multi_frame_regression(tmpdir, u, glu_ref):
    """Multiframe propkatraj call basic regression test

    Note: slow test
    """
    with tmpdir.as_cwd():
        pkatraj = propkatraj.PropkaTraj(u)
        pkatraj.run(step=10)

        # load reference data for resnum and the last frame pka
        ref_resnums, ref_pkas = pka_from_file(PSF_FRAME_NINETY_PKA)

        # test residue numbers
        resnums = pkatraj.results.pkas.columns.to_numpy()
        assert_equal(resnums, ref_resnums)

        # test final frame pka values
        assert_almost_equal(pkatraj.results.pkas.values[-1], ref_pkas, decimal=2)

        # test one data series: glu 162
        assert_almost_equal(pkatraj.results.pkas[162].values, glu_ref, decimal=2)


@pytest.mark.parametrize('start, stop, step', [
    (None, 5, None), (None, None, 25)
])
def test_start_stop_step(tmpdir, u, start, stop, step):
    """Basic test to make sure the dataframe gets populated with the right
    dimensions"""
    with tmpdir.as_cwd():
        propka = propkatraj.PropkaTraj(u)
        propka.run(start=start, stop=stop, step=step)
        pkas = propka.results.pkas

        start, stop, step = u.trajectory.check_slice_indices(start, stop, step)

        nframes = len(range(start, stop, step))

        # There should be nframes rows and 75 columns
        assert len(pkas) == nframes
        assert len(pkas.columns) == 75

        # index should be time
        assert pkas.index.name == 'time'
        times = np.array(range(start, stop, step), dtype=np.float32) + 1
        assert_almost_equal(pkas.index.values, times, decimal=5)


def test_nofilename(tmpdir):
    """See issue #23"""
    # Load a GRO file and then purposely set the filename to None
    u = mda.Universe(GRO)
    u.transfer_to_memory()
    u.trajectory.filename = None

    with tmpdir.as_cwd():
        _ = propkatraj.PropkaTraj(u).run().results.pkas
        assert os.path.isfile('current.pka')


def test_skipframe_error(tmpdir, u_badframe):
    """Test for Issue #10 - Error raised"""
    with tmpdir.as_cwd():
        with pytest.raises(RuntimeError, match="failure on frame"):
            pkatraj = propkatraj.PropkaTraj(u_badframe)
            pkatraj.run()


@pytest.mark.parametrize('top, traj, framenum', [
    (INDEXERR_FRAME14_GRO, INDEXERR_FRAME14_XTC, 14),
    (ATTERR_FRAME1_PDB, ATTERR_FRAME1_XTC, 1)
])
def test_skipframe_pass(tmpdir, caplog, top, traj, framenum):
    """Test for Issue #10 - passes, logging raise"""
    u = mda.Universe(top, traj)
    with tmpdir.as_cwd():
        wmsg = "failure on frame: {0}".format(framenum)
        with pytest.warns(UserWarning, match=wmsg):
            pkatraj = propkatraj.PropkaTraj(u, skip_failure=True)
            pkatraj.run()

        perc = 1 / u.trajectory.n_frames * 100

        msg = ("number of failed frames = 1\n"
               "percentage failure = {0}\n"
               "failed frames: {1}".format(perc, [framenum]))
        assert msg in caplog.records[0].message
        # times should have been modified in-place
        assert len(pkatraj.times) == (len(u.trajectory) - 1)
        assert u.trajectory[framenum].time not in pkatraj.times


def test_nonprotein_residues(tmpdir):
    """Test for Issue #24 - remove once PDBWriter is fixed"""
    u = mda.Universe(GRO, XTC)
    with tmpdir.as_cwd():
        wmsg = "Non protein atoms passed"
        with pytest.warns(UserWarning, match=wmsg):
            _ = propkatraj.PropkaTraj(u, select="not protein")



# propkatrj --- https://github.com/Becksteinlab/propkatraj
# Copyright (c) 2013-2020 David Dotson, Irfan Alibay, Ricky Sexton,
#                         Armin Zjajo, Shujie Fan, Oliver Beckstein
# Released under the GNU General Public License v2+

from io import StringIO
import os

import pandas as pd
import numpy as np

import propka.run as pk
import MDAnalysis as mda
from MDAnalysis.analysis.base import AnalysisBase

import warnings
import logging


class PropkaTraj(AnalysisBase):
    """Per residue pKa analysis of a trajectory.

    Runs :program:`propka` on the titrateable residues of the selected
    AtomGroup on each frame in the trajectory. Run the analysis with
    :meth:`PropkaTraj.run`, and pKa values will be stored in a
    :class:`pandas.DataFrame` named :attr:`PropkaTraj.pkas`.

    Parameters
    ----------
    atomgroup : :class:`MDAnalysis.core.universe.Universe` or :class:`MDAnalysis.core.groups.AtomGroup`
        Group of atoms containing the residues for pKa analysis. Please note
        that :class:`MDAnalysis.core.groups.UpdatingAtomGroup` are *not* supported and will
        be automatically converted to :class:`MDAnalysis.core.groups.AtomGroup`.
    select : str
        Selection string to use for selecting a subsection of atoms to use
        from the input ``atomgroup``. Note: passing non-protein residues to
        :program:`propka` may lead to incorrect results (see notes).
        [`protein`]
    skip_failure : bool
        If set to ``True``, skip frames where :program:`propka` fails. A list
        of failed frames is made available in
        :attr:`PropkaTraj.results.failed_frames_log`. If ``False`` raise a
        :exc:`RuntimeError` exception on those frames. [`False`]
        
    Attributes
    ----------
    times : np.ndarray
        times of the successfully analyzed trajectory frames
    results.pkas : pd.DataFrame
        computed pKa's for each residue as a column and each frame as a row;
        the column names are the residue numbers
    results.num_failed_frames : int
        If PROPKA failed for any frames, contains number of failed frames.
        (Needs `skip_failure` to be ``True``.)
    results.failed_frames_log : list
        Frame indices of failed frames, if `skip_failure` set to ``True``
    results.failed_times : list
        Times of failed frames.


    Notes
    -----
    Currently only the default behaviour supplemented with the `--quiet` flag
    of :program:`propka` is used.

    Temporary :program:`propka` files are written in the current working
    directory. This will leave a ``current.pka`` and ``current.propka_input``
    file. These are the temporary files for the final frame and can be removed
    safely.

    Current known issues:

    1. Due to the current behaviour of the MDAnalysis PDBWriter, non-protein
       atoms are written to PDBs using `ATOM` records instead of `HETATM`.
       This is likely to lead to undefined behaviour in :program:`propka`,
       which will likely expect `HETATM` inputs. We recommend users to only
       pass protein atoms for now. See the following issue for more details:
       https://github.com/Becksteinlab/propkatraj/issues/24
    """
    def __init__(self, atomgroup, select='protein', skip_failure=False,
                 **kwargs):
        if select is not None:
            self.ag = atomgroup.select_atoms(select).atoms
        else:
            self.ag = atomgroup.atoms

        # Issue #23 (keep until the PDBWriter is fixed)
        if len(self.ag.select_atoms('not protein')) > 0:
            wmsg = ("Non protein atoms passed to propka 3.1.\n MDAnalysis' "
                    "PDBWriter does not currently write non-standard residues "
                    "correctly as HETATM records and this may lead to "
                    "incorrect pKa predictions.\n"
                    "See https://github.com/Becksteinlab/propkatraj/issues/24 "
                    " for more details")
            warnings.warn(wmsg)

        self.tmpfile = os.path.join(os.path.curdir, 'current.pdb')
        self.skip_failure = skip_failure
        super().__init__(self.ag.universe.trajectory, **kwargs)

    def _prepare(self):
        self._pkas = []
        self.results.pkas = []
        self.results.num_failed_frames = 0
        self.results.failed_frames_log = []
        self.results.failed_times = []
        self._columns = None

    def _single_frame(self):
        pstream = mda.lib.util.NamedStream(StringIO(), self.tmpfile)
        self.ag.write(pstream)
        # reset stream for reading
        pstream.reset()

        try:
            # TODO: it would be nice to allow for other options, maybe for 3.2?
            mol = pk.single(pstream, optargs=['--quiet'])
        except (IndexError, AttributeError) as err:
            errmsg = "failure on frame: {0}".format(self._ts.frame)
            if not self.skip_failure:
                raise RuntimeError(errmsg) from err
            else:
                warnings.warn(errmsg)
                self.results.num_failed_frames += 1
                self.results.failed_frames_log.append(self._ts.frame)
                self.results.failed_times.append(self._ts.time)
        else:
            confname = mol.conformation_names[0]
            conformation = mol.conformations[confname]
            groups = conformation.get_titratable_groups()

            # extract pka estimates from each residue
            self._pkas.append([g.pka_value for g in groups])
            if self._columns is None:
                self._columns = [g.atom.resNumb for g in groups]
        finally:
            # deallocate stream
            pstream.close(force=True)

    def _conclude(self):
        # Ouput failed frames
        if self.results.num_failed_frames > 0:
            perc_failure = (self.results.num_failed_frames / self.n_frames) * 100

            # if frames have failed we need to ammend times accordingly
            failed_indices = [np.where(self.times == i) for i in
                              self.results.failed_times]
            self.times = np.delete(self.times, failed_indices)

            wmsg = ("number of failed frames = {0}\n"
                    "percentage failure = {1}\n"
                    "failed frames: {2}".format(self.results.num_failed_frames,
                                                perc_failure,
                                                self.results.failed_frames_log))
            logging.warning(wmsg)

        self.results.pkas = pd.DataFrame(self._pkas, index=pd.Index(self.times,
                                         name='time', dtype='float64'),
                                         columns=self._columns)
